"""Run Bayesian analysis of data from an Ecological Momentary Assessment (EMA) study.
This script should be used as a template,
to be copied and modified for the desired analysis.

*** Usage, four main steps, see also explicit template example below

*1: Create an EmaFrame instance to define experiment and select input data.

*2: Load a set of collected EMA data into an EmaDataSet instance

*3: Initialize and train a Bayesian model corresponding to observed data.

*4: Display results and save figures and tables to a directory tree


*** Version history:
2021-11-22, first functional version
"""
import numpy as np
from pathlib import Path
import logging
import pickle

from EmaCalc.ema_simulation import EmaSimPopulation, EmaSimExperiment
from EmaCalc.ema_simulation import SubjectThurstone, SubjectBradley

from EmaCalc.ema_data import EmaFrame, EmaDataSet
from EmaCalc.ema_model import EmaModel
from EmaCalc.ema_display import EmaDisplaySet
from EmaCalc import ema_logging


# ------------------------ Set up working directory and result logging:
work_path = Path.home() / 'Documents' / 'EMA_sim'  # or whatever...
data_path = work_path / 'data'  # to use simulation data generated by run_sim.py
result_path = work_path / 'result'  # or whatever

ema_logging.setup(save_path=work_path,
                  log_file='run_sim_log.txt')  # to save the log file
# OR
# ema_logging.setup()  # to get only console output
model_file = 'test_ema_model.pkl'  # name of saved model file (if saved)

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


# --------------- Define Experimental Framework: Scenarios, Attributes, and Ratings

# NOTE: This example refers to data generated by script run_sim.py
# Edit as needed for any other EMA study

sim_scenarios = {'Stage': ('',),  # only ONE Test Stage with empty label
                 'HA': ('A', 'B'),  # Two Hearing-aid programs
                 'CoSS': [f'{i}.' for i in range(1, 8)],    # Seven CoSS categories
                 }  # nominal variables, same for all (Sub-)Populations
# NOTE: scenario dimension 'Stage' may be omitted, if only one Stage category

emf = EmaFrame(scenarios=sim_scenarios,
               stage_key='Stage',
               attribute_grades={'Speech': ['Very Hard',
                                            'Hard',
                                            'Easy',
                                            'Very Easy',
                                            'Perfect']},  # ordinal rating categories
               )

# --------------- Load data from previously saved xlsx file(s):

# NOTE: if 'grouping' is defined, group directory names must match given labels
# With template example: data_path / 'Age_old' is top of directory tree with data files.
# If several 'grouping' factors are defined, the directory three structure must match the factors.
# If no grouping is defined: data_path is top of directory tree with all data files

ds = EmaDataSet.load(emf, data_path,
                     grouping={'Age': ('old',),  # OR 'Age': ('old', 'young'),
                               },  # may be omitted if only one group
                     subject='sheet',   # subject ID defined in xlsx sheet title
                     ema_vars={'HA': 'B',   # HA category in column B, etc.
                               'CoSS': 'C',
                               'Speech': 'F'
                               }
                     )

logger.info(f'Using data ds=\n{ds}')

# ------------------------------- Learn Analysis Model from loaded data set:

# Model ordinal-regression effects of Scenarios on each Attribute:
regression_effects = [  # 'HA', 'CoSS',  # main linear regression effects, no interactions
                      ('HA', 'CoSS')  # both main AND interaction effects
                      # 'Stage',  # only if there are several Stage categories
                      ]
# NOTE: regression effects may include any combination of scenario dimensions, but
# including all interactions -> many free model parameters,
# possibly -> less precise estimation for each parameter.
# In this example: 'HA', 'CoSS' -> 2 + 7 - 1 = 8 regression-effect parameters
#                ('HA', 'CoSS') -> 2 * 7 - 1 = 13 regression-effect parameters

emm = EmaModel.initialize(ds, regression_effects,
                          max_n_comp=10)
# max_n_comp = max number of mixture components in population model

ll = emm.learn(max_hours=1.)
logger.info('ll= ' + np.array2string(np.array(ll), precision=5))
emm.prune()  # keep only active mixture components

# -------- Save learned EmaModel (optional):
with (work_path / model_file).open('wb') as f:
    pickle.dump(emm, f)

# ------------------------------- generate all result displays:

emd = EmaDisplaySet.display(emm,
                            scenarios=['CoSS',  # CoSS profile, aggregated across HA
                                       ('CoSS', 'HA'),  # CoSS profiles, conditional on HA
                                       ('HA', 'CoSS'),  # HA profiles, conditional on CoSS
                                       ],
                            attributes=[('Speech', 'CoSS'),     # Speech, main effect of CoSS
                                        ('Speech', 'HA'),       # Speech, main effect of HA
                                        ('Speech', ('CoSS', 'HA')),     # joint effect of both
                                        ],
                            population_individual=True,  # random individual in the population
                            population_mean=True,  # population mean
                            subjects=True,  # individual results -> MANY plots and tables
                            show_intervals=True,  # response thresholds in attribute plots
                            percentiles=[5, 25, 50, 75, 95],  # in profile plots and tables
                            credibility_limit=0.6,  # minimum credibility for differences
                            figure_format='pdf',  # or 'eps', 'jpg'
                            table_format='latex',  # or 'tab' for tab-limited text style
                            )


# ------------------------------- save all result displays in directory tree
emd.save(result_path)

logging.info(f'All results saved in {result_path}')

logging.shutdown()
